<!DOCTYPE html>
<html>
  <head>
    <title>Admin Account Created</title>
  </head>
  <body>
    <h1>Admin account succesfully created</h1>
    <p>Username: {{ username }}</p>
    <p>Please visit the webapp via /app</p>
  </body>
</html>
